@@ -2,6 +2,11 @@
 # include <blkid.h>
 #endif
 
+#include "blkdev.h"
+#ifdef __linux__
+# include "partx.h"
+#endif
+#include "loopdev.h"
 #include "fdiskP.h"
 
 
@@ -47,6 +52,8 @@ struct fdisk_context *fdisk_new_context(void)
 	cxt->dev_fd = -1;
 	cxt->refcount = 1;
 
+	INIT_LIST_HEAD(&cxt->wipes);
+
 	/*
 	 * Allocate label specific structs.
 	 *
@@ -59,6 +66,8 @@ struct fdisk_context *fdisk_new_context(void)
 	cxt->labels[ cxt->nlabels++ ] = fdisk_new_sgi_label(cxt);
 	cxt->labels[ cxt->nlabels++ ] = fdisk_new_sun_label(cxt);
 
+	bindtextdomain(LIBFDISK_TEXTDOMAIN, LOCALEDIR);
+
 	return cxt;
 }
 
@@ -94,12 +103,13 @@ static int init_nested_from_parent(struct fdisk_context *cxt, int isnew)
 	cxt->user_log_sector =	parent->user_log_sector;
 	cxt->user_pyh_sector =  parent->user_pyh_sector;
 
-	/* parent <--> nested independent setting, initialize for new nested 
+	/* parent <--> nested independent setting, initialize for new nested
 	 * contexts only */
 	if (isnew) {
 		cxt->listonly =	parent->listonly;
 		cxt->display_details =	parent->display_details;
 		cxt->display_in_cyl_units = parent->display_in_cyl_units;
+		cxt->protect_bootbits = parent->protect_bootbits;
 	}
 
 	free(cxt->dev_path);
@@ -111,6 +121,8 @@ static int init_nested_from_parent(struct fdisk_context *cxt, int isnew)
 			return -ENOMEM;
 	}
 
+	INIT_LIST_HEAD(&cxt->wipes);
+
 	return 0;
 }
 
@@ -201,7 +213,7 @@ void fdisk_ref_context(struct fdisk_context *cxt)
  * If no @name specified then returns the current context label.
  *
  * The label is allocated and maintained within the context #cxt. There is
- * nothing like reference counting for labels, you cannot delallocate the
+ * nothing like reference counting for labels, you cannot deallocate the
  * label.
  *
  * Returns: label struct or NULL in case of error.
@@ -303,6 +315,137 @@ int fdisk_has_label(struct fdisk_context *cxt)
 	return cxt && cxt->label;
 }
 
+/**
+ * fdisk_has_protected_bootbits:
+ * @cxt: fdisk context
+ *
+ * Returns: return 1 if boot bits protection enabled.
+ */
+int fdisk_has_protected_bootbits(struct fdisk_context *cxt)
+{
+	return cxt && cxt->protect_bootbits;
+}
+
+/**
+ * fdisk_enable_bootbits_protection:
+ * @cxt: fdisk context
+ * @enable: 1 or 0
+ *
+ * The library zeroizes all the first sector when create a new disk label by
+ * default.  This function allows to control this behavior. For now it's
+ * supported for MBR and GPT.
+ *
+ * Returns: 0 on success, < 0 on error.
+ */
+int fdisk_enable_bootbits_protection(struct fdisk_context *cxt, int enable)
+{
+	if (!cxt)
+		return -EINVAL;
+	cxt->protect_bootbits = enable ? 1 : 0;
+	return 0;
+}
+/**
+ * fdisk_disable_dialogs
+ * @cxt: fdisk context
+ * @enable: 1 or 0
+ *
+ * The library uses dialog driven partitioning by default.
+ *
+ * Returns: 0 on success, < 0 on error.
+ *
+ * Since: 2.31
+ */
+int fdisk_disable_dialogs(struct fdisk_context *cxt, int disable)
+{
+	if (!cxt)
+		return -EINVAL;
+
+	cxt->no_disalogs = disable;
+	return 0;
+}
+
+/**
+ * fdisk_has_dialogs
+ * @cxt: fdisk context
+ *
+ * See fdisk_disable_dialogs()
+ *
+ * Returns: 1 if dialog driven partitioning enabled (default), or 0.
+ *
+ * Since: 2.31
+ */
+int fdisk_has_dialogs(struct fdisk_context *cxt)
+{
+	return cxt->no_disalogs == 0;
+}
+
+/**
+ * fdisk_enable_wipe
+ * @cxt: fdisk context
+ * @enable: 1 or 0
+ *
+ * The library removes all PT/filesystem/RAID signatures before it writes
+ * partition table. The probing area where it looks for signatures is from
+ * the begin of the disk. The device is wiped by libblkid.
+ *
+ * See also fdisk_wipe_partition().
+ *
+ * Returns: 0 on success, < 0 on error.
+ */
+int fdisk_enable_wipe(struct fdisk_context *cxt, int enable)
+{
+	if (!cxt)
+		return -EINVAL;
+
+	fdisk_set_wipe_area(cxt, 0, cxt->total_sectors, enable);
+	return 0;
+}
+
+/**
+ * fdisk_has_wipe
+ * @cxt: fdisk context
+ *
+ * Returns the current wipe setting. See fdisk_enable_wipe().
+ *
+ * Returns: 0 on success, < 0 on error.
+ */
+int fdisk_has_wipe(struct fdisk_context *cxt)
+{
+	if (!cxt)
+		return 0;
+
+	return fdisk_has_wipe_area(cxt, 0, cxt->total_sectors);
+}
+
+
+/**
+ * fdisk_get_collision
+ * @cxt: fdisk context
+ *
+ * Returns: name of the filesystem or RAID detected on the device or NULL.
+ */
+const char *fdisk_get_collision(struct fdisk_context *cxt)
+{
+	return cxt->collision;
+}
+
+/**
+ * fdisk_is_ptcollision:
+ * @cxt: fdisk context
+ *
+ * The collision detected by libblkid (usally another partition table). Note
+ * that libfdisk does not support all partitions tables, so fdisk_has_label()
+ * may return false, but fdisk_is_ptcollision() may return true.
+ *
+ * Since: 2.30
+ *
+ * Returns: 0 or 1
+ */
+int fdisk_is_ptcollision(struct fdisk_context *cxt)
+{
+	return cxt->pt_collision;
+}
+
 /**
  * fdisk_get_npartitions:
  * @cxt: context
@@ -357,7 +500,7 @@ int fdisk_is_labeltype(struct fdisk_context *cxt, enum fdisk_labeltype id)
 {
 	assert(cxt);
 
-	return cxt->label && fdisk_label_get_type(cxt->label) == id;
+	return cxt->label && (unsigned)fdisk_label_get_type(cxt->label) == id;
 }
 
 /**
@@ -396,6 +539,11 @@ static void reset_context(struct fdisk_context *cxt)
 	free(cxt->dev_path);
 	cxt->dev_path = NULL;
 
+	free(cxt->collision);
+	cxt->collision = NULL;
+
+	memset(&cxt->dev_st, 0, sizeof(cxt->dev_st));
+
 	cxt->dev_fd = -1;
 	cxt->firstsector = NULL;
 	cxt->firstsector_bufsz = 0;
@@ -406,59 +554,8 @@ static void reset_context(struct fdisk_context *cxt)
 	cxt->script = NULL;
 
 	cxt->label = NULL;
-}
 
-/*
- * This function prints a warning if the device is not wiped (e.g. wipefs(8).
- * Please don't call this function if there is already a PT.
- *
- * Returns: 0 if nothing found, < 0 on error, 1 if found a signature
- */
-static int warn_wipe(struct fdisk_context *cxt)
-{
-#ifdef HAVE_LIBBLKID
-	blkid_probe pr;
-#endif
-	int rc = 0;
-
-	assert(cxt);
-
-	if (fdisk_has_label(cxt) || cxt->dev_fd < 0)
-		return -EINVAL;
-#ifdef HAVE_LIBBLKID
-	DBG(CXT, ul_debugobj(cxt, "wipe check: initialize libblkid prober"));
-
-	pr = blkid_new_probe();
-	if (!pr)
-		return -ENOMEM;
-	rc = blkid_probe_set_device(pr, cxt->dev_fd, 0, 0);
-	if (rc)
-		return rc;
-
-	blkid_probe_enable_superblocks(pr, 1);
-	blkid_probe_set_superblocks_flags(pr, BLKID_SUBLKS_TYPE);
-	blkid_probe_enable_partitions(pr, 1);
-
-	/* we care about the first found FS/raid, so don't call blkid_do_probe()
-	 * in loop or don't use blkid_do_fullprobe() ... */
-	rc = blkid_do_probe(pr);
-	if (rc == 0) {
-		const char *name = NULL;
-
-		if (blkid_probe_lookup_value(pr, "TYPE", &name, 0) == 0 ||
-		    blkid_probe_lookup_value(pr, "PTTYPE", &name, 0) == 0) {
-			fdisk_warnx(cxt, _(
-				"%s: device contains a valid '%s' signature; it is "
-				"strongly recommended to wipe the device with "
-				"wipefs(8) if this is unexpected, in order to "
-				"avoid possible collisions"), cxt->dev_path, name);
-			rc = 1;
-		}
-	}
-
-	blkid_free_probe(pr);
-#endif
-	return rc;
+	fdisk_free_wipe_areas(cxt);
 }
 
 /**
@@ -491,7 +588,7 @@ int fdisk_assign_device(struct fdisk_context *cxt,
 		int rc, org = fdisk_is_listonly(cxt->parent);
 
 		/* assign_device() is sensitive to "listonly" mode, so let's
-		 * follow the current context setting for the parent to avoid 
+		 * follow the current context setting for the parent to avoid
 		 * unwanted extra warnings. */
 		fdisk_enable_listonly(cxt->parent, fdisk_is_listonly(cxt));
 
@@ -509,7 +606,10 @@ int fdisk_assign_device(struct fdisk_context *cxt,
 
 	fd = open(fname, (readonly ? O_RDONLY : O_RDWR ) | O_CLOEXEC);
 	if (fd < 0)
-		return -errno;
+		goto fail;
+
+	if (fstat(fd, &cxt->dev_st) != 0)
+		goto fail;
 
 	cxt->readonly = readonly;
 	cxt->dev_fd = fd;
@@ -520,28 +620,32 @@ int fdisk_assign_device(struct fdisk_context *cxt,
 	fdisk_discover_topology(cxt);
 	fdisk_discover_geometry(cxt);
 
+	fdisk_apply_user_device_properties(cxt);
+
 	if (fdisk_read_firstsector(cxt) < 0)
 		goto fail;
 
-	/* detect labels and apply labes specific stuff (e.g geomery)
-	 * to the context */
 	fdisk_probe_labels(cxt);
 
-	/* let's apply user geometry *after* label prober
-	 * to make it possible to override in-label setting */
-	fdisk_apply_user_device_properties(cxt);
+	fdisk_apply_label_device_properties(cxt);
 
 	/* warn about obsolete stuff on the device if we aren't in
 	 * list-only mode and there is not PT yet */
-	if (!fdisk_is_listonly(cxt) && !fdisk_has_label(cxt))
-		warn_wipe(cxt);
+	if (!fdisk_is_listonly(cxt) && !fdisk_has_label(cxt)
+	    && fdisk_check_collisions(cxt) < 0)
+		goto fail;
 
 	DBG(CXT, ul_debugobj(cxt, "initialized for %s [%s]",
 			      fname, readonly ? "READ-ONLY" : "READ-WRITE"));
 	return 0;
 fail:
-	DBG(CXT, ul_debugobj(cxt, "failed to assign device"));
-	return -errno;
+	{
+		int rc = -errno;
+		if (fd >= 0)
+			close(fd);
+		DBG(CXT, ul_debugobj(cxt, "failed to assign device [rc=%d]", rc));
+		return rc;
+	}
 }
 
 /**
@@ -567,6 +671,8 @@ int fdisk_deassign_device(struct fdisk_context *cxt, int nosync)
 		return rc;
 	}
 
+	DBG(CXT, ul_debugobj(cxt, "de-assigning device %s", cxt->dev_path));
+
 	if (cxt->readonly)
 		close(cxt->dev_fd);
 	else {
@@ -590,6 +696,227 @@ int fdisk_deassign_device(struct fdisk_context *cxt, int nosync)
 	return 0;
 }
 
+/**
+ * fdisk_reassign_device:
+ * @cxt: context
+ *
+ * This function is "hard reset" of the context and it does not write anything
+ * to the device. All in-memory changes associated with the context will be
+ * lost. It's recommended to use this function after some fatal problem when the
+ * context (and label specific driver) is in an undefined state.
+ *
+ * Returns: 0 on success, < 0 on error.
+ */
+int fdisk_reassign_device(struct fdisk_context *cxt)
+{
+	char *devname;
+	int rdonly, rc;
+
+	assert(cxt);
+	assert(cxt->dev_fd >= 0);
+
+	DBG(CXT, ul_debugobj(cxt, "re-assigning device %s", cxt->dev_path));
+
+	devname = strdup(cxt->dev_path);
+	if (!devname)
+		return -ENOMEM;
+
+	rdonly = cxt->readonly;
+
+	fdisk_deassign_device(cxt, 1);
+	rc = fdisk_assign_device(cxt, devname, rdonly);
+	free(devname);
+
+	return rc;
+}
+
+/**
+ * fdisk_reread_partition_table:
+ * @cxt: context
+ *
+ * Force *kernel* to re-read partition table on block devices.
+ *
+ * Returns: 0 on success, < 0 in case of error.
+ */
+int fdisk_reread_partition_table(struct fdisk_context *cxt)
+{
+	int i = 0;
+
+	assert(cxt);
+	assert(cxt->dev_fd >= 0);
+
+	if (!S_ISBLK(cxt->dev_st.st_mode))
+		return 0;
+	else {
+		DBG(CXT, ul_debugobj(cxt, "calling re-read ioctl"));
+		sync();
+#ifdef BLKRRPART
+		fdisk_info(cxt, _("Calling ioctl() to re-read partition table."));
+		i = ioctl(cxt->dev_fd, BLKRRPART);
+#else
+		errno = ENOSYS;
+		i = 1;
+#endif
+	}
+
+	if (i) {
+		fdisk_warn(cxt, _("Re-reading the partition table failed."));
+		fdisk_info(cxt,	_(
+			"The kernel still uses the old table. The "
+			"new table will be used at the next reboot "
+			"or after you run partprobe(8) or kpartx(8)."));
+		return -errno;
+	}
+
+	return 0;
+}
+
+#ifdef __linux__
+static inline int add_to_partitions_array(
+			struct fdisk_partition ***ary,
+			struct fdisk_partition *pa,
+			size_t *n, size_t nmax)
+{
+	if (!*ary) {
+		*ary = calloc(nmax, sizeof(struct fdisk_partition *));
+		if (!*ary)
+			return -ENOMEM;
+	}
+	(*ary)[*n] = pa;
+	(*n)++;
+	return 0;
+}
+#endif
+
+/**
+ * fdisk_reread_changes:
+ * @cxt: context
+ * @org: original layout (on disk)
+ *
+ * Like fdisk_reread_partition_table() but don't forces kernel re-read all
+ * partition table. The BLKPG_* ioctls are used for individual partitions. The
+ * advantage is that unmodified partitions maybe mounted.
+ *
+ * The function behavies like fdisk_reread_partition_table() on systems where
+ * are no avaialble BLKPG_* ioctls.
+ *
+ * Returns: <0 on error, or 0.
+ */
+#ifdef __linux__
+int fdisk_reread_changes(struct fdisk_context *cxt, struct fdisk_table *org)
+{
+	struct fdisk_table *tb = NULL;
+	struct fdisk_iter itr;
+	struct fdisk_partition *pa;
+	struct fdisk_partition **rem = NULL, **add = NULL, **upd = NULL;
+	int change, rc = 0, err = 0;
+	size_t nparts, i, nadds = 0, nupds = 0, nrems = 0;
+
+	DBG(CXT, ul_debugobj(cxt, "rereading changes"));
+
+	fdisk_reset_iter(&itr, FDISK_ITER_FORWARD);
+
+	/* the current layout */
+	fdisk_get_partitions(cxt, &tb);
+	/* maximal number of partitions */
+	nparts = max(fdisk_table_get_nents(tb), fdisk_table_get_nents(org));
+
+	while (fdisk_diff_tables(org, tb, &itr, &pa, &change) == 0) {
+		if (change == FDISK_DIFF_UNCHANGED)
+			continue;
+		switch (change) {
+		case FDISK_DIFF_REMOVED:
+			rc = add_to_partitions_array(&rem, pa, &nrems, nparts);
+			break;
+		case FDISK_DIFF_ADDED:
+			rc = add_to_partitions_array(&add, pa, &nadds, nparts);
+			break;
+		case FDISK_DIFF_RESIZED:
+			rc = add_to_partitions_array(&upd, pa, &nupds, nparts);
+			break;
+		case FDISK_DIFF_MOVED:
+			rc = add_to_partitions_array(&rem, pa, &nrems, nparts);
+			rc = add_to_partitions_array(&add, pa, &nadds, nparts);
+			break;
+		}
+		if (rc != 0)
+			goto done;
+	}
+
+	for (i = 0; i < nrems; i++) {
+		pa = rem[i];
+		DBG(PART, ul_debugobj(pa, "#%zu calling BLKPG_DEL_PARTITION", pa->partno));
+		if (partx_del_partition(cxt->dev_fd, pa->partno + 1) != 0) {
+			fdisk_warn(cxt, _("Failed to remove partition %zu from system"), pa->partno + 1);
+			err++;
+		}
+	}
+	for (i = 0; i < nupds; i++) {
+		pa = upd[i];
+		DBG(PART, ul_debugobj(pa, "#%zu calling BLKPG_RESIZE_PARTITION", pa->partno));
+		if (partx_resize_partition(cxt->dev_fd, pa->partno + 1, pa->start, pa->size) != 0) {
+			fdisk_warn(cxt, _("Failed to update system information about partition %zu"), pa->partno + 1);
+			err++;
+		}
+	}
+	for (i = 0; i < nadds; i++) {
+		pa = add[i];
+		DBG(PART, ul_debugobj(pa, "#%zu calling BLKPG_ADD_PARTITION", pa->partno));
+		if (partx_add_partition(cxt->dev_fd, pa->partno + 1, pa->start, pa->size) != 0) {
+			fdisk_warn(cxt, _("Failed to add partition %zu to system"), pa->partno + 1);
+			err++;
+		}
+	}
+	if (err)
+		fdisk_info(cxt,	_(
+			"The kernel still uses the old partitions. The new "
+			"table will be used at the next reboot. "));
+done:
+	free(rem);
+	free(add);
+	free(upd);
+	fdisk_unref_table(tb);
+	return rc;
+}
+#else
+int fdisk_reread_changes(struct fdisk_context *cxt,
+			 struct fdisk_table *org __attribute__((__unused__))) {
+	return fdisk_reread_partition_table(cxt);
+}
+#endif
+
+/**
+ * fdisk_device_is_used:
+ * @cxt: context
+ *
+ * On systems where is no BLKRRPART ioctl the function returns zero and
+ * sets errno to ENOSYS.
+ *
+ * Returns: 1 if the device assigned to the context is used by system, or 0.
+ */
+int fdisk_device_is_used(struct fdisk_context *cxt)
+{
+	int rc = 0;
+
+	assert(cxt);
+	assert(cxt->dev_fd >= 0);
+
+	errno = 0;
+
+#ifdef BLKRRPART
+	/* it seems kernel always return EINVAL for BLKRRPART on loopdevices */
+	if (S_ISBLK(cxt->dev_st.st_mode)
+	    && major(cxt->dev_st.st_rdev) != LOOPDEV_MAJOR) {
+		DBG(CXT, ul_debugobj(cxt, "calling re-read ioctl"));
+		rc = ioctl(cxt->dev_fd, BLKRRPART) != 0;
+	}
+#else
+	errno = ENOSYS;
+#endif
+	DBG(CXT, ul_debugobj(cxt, "device used: %s [errno=%d]", rc ? "TRUE" : "FALSE", errno));
+	return rc;
+}
+
 /**
  * fdisk_is_readonly:
  * @cxt: context
@@ -602,6 +929,20 @@ int fdisk_is_readonly(struct fdisk_context *cxt)
 	return cxt->readonly;
 }
 
+/**
+ * fdisk_is_regfile:
+ * @cxt: context
+ *
+ * Since: 2.30
+ *
+ * Returns: 1 if open file descriptor is regular file rather than a block device.
+ */
+int fdisk_is_regfile(struct fdisk_context *cxt)
+{
+	assert(cxt);
+	return S_ISREG(cxt->dev_st.st_mode);
+}
+
 /**
  * fdisk_unref_context:
  * @cxt: fdisk context
@@ -610,7 +951,7 @@ int fdisk_is_readonly(struct fdisk_context *cxt)
  */
 void fdisk_unref_context(struct fdisk_context *cxt)
 {
-	int i;
+	unsigned i;
 
 	if (!cxt)
 		return;
@@ -642,7 +983,7 @@ void fdisk_unref_context(struct fdisk_context *cxt)
 /**
  * fdisk_enable_details:
  * @cxt: context
- * @enable: true/flase
+ * @enable: true/false
  *
  * Enables or disables "details" display mode. This function has effect to
  * fdisk_partition_to_string() function.
@@ -671,7 +1012,7 @@ int fdisk_is_details(struct fdisk_context *cxt)
 /**
  * fdisk_enable_listonly:
  * @cxt: context
- * @enable: true/flase
+ * @enable: true/false
  *
  * Just list partition only, don't care about another details, mistakes, ...
  *
@@ -705,7 +1046,7 @@ int fdisk_is_listonly(struct fdisk_context *cxt)
  * This is pure shit, unfortunately for example Sun addresses begin of the
  * partition by cylinders...
  *
- * Returns: 0 on succes, <0 on error.
+ * Returns: 0 on success, <0 on error.
  */
 int fdisk_set_unit(struct fdisk_context *cxt, const char *str)
 {
@@ -871,7 +1212,7 @@ fdisk_sector_t fdisk_get_first_lba(struct fdisk_context *cxt)
  * @lba: first possible logical sector for data
  *
  * It's strongly recommended to use the default library setting. The first LBA
- * is always reseted by fdisk_assign_device(), fdisk_override_geometry()
+ * is always reset by fdisk_assign_device(), fdisk_override_geometry()
  * and fdisk_reset_alignment(). This is very low level function and library
  * does not check if your setting makes any sense.
  *
@@ -909,7 +1250,7 @@ fdisk_sector_t fdisk_get_last_lba(struct fdisk_context *cxt)
  * @lba: last possible logical sector
  *
  * It's strongly recommended to use the default library setting. The last LBA
- * is always reseted by fdisk_assign_device(), fdisk_override_geometry() and
+ * is always reset by fdisk_assign_device(), fdisk_override_geometry() and
  * fdisk_reset_alignment().
  *
  * The default is number of sectors on the device, but maybe modified by the
@@ -922,7 +1263,7 @@ fdisk_sector_t fdisk_set_last_lba(struct fdisk_context *cxt, fdisk_sector_t lba)
 {
 	assert(cxt);
 
-	if (lba > cxt->total_sectors - 1 && lba < 1)
+	if (lba > cxt->total_sectors - 1 || lba < 1)
 		return -ERANGE;
 	cxt->last_lba = lba;
 	return 0;
@@ -952,7 +1293,7 @@ int fdisk_set_size_unit(struct fdisk_context *cxt, int unit)
  *
  * Returns: unit
  */
-int fdisk_get_size_units(struct fdisk_context *cxt)
+int fdisk_get_size_unit(struct fdisk_context *cxt)
 {
 	assert(cxt);
 	return cxt->sizeunit;
@@ -986,7 +1327,7 @@ const char *fdisk_get_devname(struct fdisk_context *cxt)
  * fdisk_get_devfd:
  * @cxt: context
  *
- * Retruns: device file descriptor.
+ * Returns: device file descriptor.
  */
 int fdisk_get_devfd(struct fdisk_context *cxt)
 {
@@ -1034,8 +1375,6 @@ int fdisk_missing_geometry(struct fdisk_context *cxt)
 {
 	int rc;
 
-	assert(cxt);
-
 	if (!cxt || !cxt->label)
 		return 0;
 
